@@ -1,7 +1,9 @@
 """This module contains the InstaClient class"""
+from logging import log
 from os import waitpid
 from selenium import webdriver
 from selenium.webdriver.remote.webelement import WebElement
+from selenium.webdriver.support.expected_conditions import presence_of_element_located
 from selenium.webdriver.support.wait import WebDriverWait
 from selenium.webdriver.common.by import By
 from selenium.webdriver.support import expected_conditions as EC
@@ -18,7 +20,7 @@ class InstaClient:
     CHROMEDRIVER=1
     LOCAHOST=1
     WEB_SERVER=2
-    def __init__(self, driver_type: int=CHROMEDRIVER, host:int=LOCAHOST):
+    def __init__(self, driver_type: int=CHROMEDRIVER, host_type:int=LOCAHOST):
         """
         Create an `InstaClient` object to access the instagram website.
 
@@ -31,56 +33,41 @@ class InstaClient:
             InvaildDriverError: Raised if driver int does not correspond to any driver type.
             error: Normal Exception, raised if anything fails when creating the client.
         """
-        try:
-            if driver_type == self.CHROMEDRIVER:
-                if host == self.WEB_SERVER:
-                    # Running on web server
-                    chrome_options = webdriver.ChromeOptions()
-                    chrome_options.binary_location = os.environ.get("GOOGLE_CHROME_BIN")
-                    chrome_options.add_argument('--user-agent=Mozilla/5.0 (iPhone; CPU iPhone OS 10_3 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) CriOS/56.0.2924.75 Mobile/14E5239e Safari/602.1')
-                    chrome_options.add_argument("--window-size=343,915")
-                    chrome_options.add_argument("--headless")
-                    chrome_options.add_argument("--disable-dev-shm-usage")
-                    chrome_options.add_argument("--no-sandbox")
-                    self.driver = webdriver.Chrome(executable_path=os.environ.get("CHROMEDRIVER_PATH"), chrome_options=chrome_options)
-                elif host == self.LOCAHOST:
-                    # Running locally
-                    chrome_options = webdriver.ChromeOptions()
-                    chrome_options.add_argument('--user-agent=Mozilla/5.0 (iPhone; CPU iPhone OS 10_3 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) CriOS/56.0.2924.75 Mobile/14E5239e Safari/602.1')
-                    chrome_options.add_argument("--window-size=343,915")
-                    #chrome_options.add_argument("--headless")
-                    chrome_options.add_argument("--disable-dev-shm-usage")
-                    chrome_options.add_argument("--no-sandbox")
-                    self.driver = webdriver.Chrome(executable_path='instaclient/drivers/chromedriver.exe', chrome_options=chrome_options)
-                else:
-                    raise InvaildHostError(host)
-            else:
-                raise InvaildDriverError(driver_type)
-        except Exception as error:
-            raise error
+        self.driver_type = driver_type
+        self.host_type = host_type
         self.logged_in = False
+        self.driver = None
+        self.username = None
+        self.password = None
+        self.__init_driver()
 
 
     @insta_method
-    def check_status(self):
+    def check_status(self, discard_driver:bool=False):
         """
         Check if account is currently logged in. Returns True if account is logged in. Sets the `instaclient.logged_in` variable accordingly.
-
+        Returns False if the driver is not open yet - even if the Instagram credentials (`username` and `password`) are correct.
 
         Returns:
-            bool: True if client is logged in, False if client is not connected.
+            bool: True if client is logged in, False if client is not connected or webdriver is not open.
         """
+        if not self.driver:
+            return False
+        self.driver.get(ClientUrls.HOME_URL)
         icon = self.__check_existence(EC.presence_of_element_located((By.XPATH, Paths.NAV_BAR)), wait_time=4)
         if icon:
             self.logged_in = True
-            return True
+            result = True
         else:
             self.logged_in = False
-            return False
+            result = False
+        if discard_driver:
+            self.__discard_driver()
+        return result
 
 
     @insta_method
-    def login(self, username:str, password:str, check_user:bool=True):
+    def login(self, username:str, password:str, check_user:bool=True, discard_driver:bool=False):
         """
         Sign Into Instagram with credentials. Go through 2FA if necessary. Sets the InstaClient variable `InstaClient.logged_in` to True if login was successful.
 
@@ -99,6 +86,11 @@ class InstaClient:
         """
         self.username = username
         self.password = password
+        
+        # Initiate Driver
+        if not self.driver:
+            self.__init_driver()
+
         # Get Elements
         try:
             # Attempt Login
@@ -127,6 +119,8 @@ class InstaClient:
         except Exception as error:
             # User already logged in ?
             result = self.check_status()
+            if discard_driver:
+                self.__discard_driver()
             if not result:
                 raise error
             else:
@@ -177,16 +171,20 @@ class InstaClient:
         else:
             self.logged_in = True
 
-        # Detect and dismiss save info Dialog
-        self.driver.get(ClientUrls.HOME_URL)
+        # Discard Driver or complete login
+        if discard_driver:
+            self.__discard_driver()
+        else:
+            # Detect and dismiss save info Dialog
+            self.driver.get(ClientUrls.HOME_URL)
 
-        # Detect 'Turn On Notifications' Box
-        try:
-            no_notifications_btn = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.NO_NOTIFICATIONS_BTN)), wait_time=3, url=ClientUrls.HOME_URL)
-            no_notifications_btn.click()
-        except:
-            pass
-        self.dismiss_dialogue()
+            # Detect 'Turn On Notifications' Box
+            try:
+                no_notifications_btn = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.NO_NOTIFICATIONS_BTN)), wait_time=3, url=ClientUrls.HOME_URL)
+                no_notifications_btn.click()
+            except:
+                pass
+            self.dismiss_dialogue()
         return self.logged_in
 
 
@@ -225,7 +223,7 @@ class InstaClient:
 
 
     @insta_method
-    def input_security_code(self, code:int or str):
+    def input_security_code(self, code:int or str, discard_driver:bool=False):
         """
         Complete login procedure started with `InstaClient.login()` and insert security code required if `instaclient.errors.common.SuspiciousLoginAttemptError` is raised. Sets `InstaClient.logged_in` attribute to True if login was successful.
 
@@ -258,12 +256,15 @@ class InstaClient:
             raise InvalidSecurityCodeError()
 
         self.logged_in = True
-        self.dismiss_dialogue()
+        if discard_driver:
+            self.__discard_driver()
+        else:
+            self.dismiss_dialogue()
         return self.logged_in
 
 
     @insta_method
-    def input_verification_code(self, code:int or str):
+    def input_verification_code(self, code:int or str, discard_driver:bool=False):
         """
         Complete login procedure started with `InstaClient.login()` and insert 2FA security code. Sets `instaclient.logged_in` to True if login was successful.
 
@@ -292,35 +293,68 @@ class InstaClient:
             # Auth Correct
             self.logged_in = True
             self.dismiss_dialogue()
+            if discard_driver:
+                self.__discard_driver()
             return self.logged_in
 
 
     @insta_method
-    def follow_user(self, user:str):
+    def follow_user(self, user:str, nav_to_user:bool=True, discard_driver:bool=False):
         """
-        Follows user(s)
+        follow_user follows the instagram user that matches the username in the `user` attribute.
+        If the target account is private, a follow request will be sent to such user and a `PrivateAccountError` will be raised.
 
         Args:
-            user:str: Username of the user to follow
+            user (str): Username of the user to follow.
+            discard_driver (bool, optional): If set to True, the driver will be discarded at the end of the method. Defaults to False.
+
+        Raises:
+            PrivateAccountError: Raised if the `user` is a private account - A request to follow the user will be sent eitherway.
+            InvalidUserError: Raised if the `user` is invalid
         """
+        if not self.driver:
+            self.__init_driver(login=True)
+        
+        # Navigate to User Page
+        if nav_to_user:
+            self.nav_user(user, check_user=False)
+        
+        # Check User Vadility
+        try:
+            result = self.is_valid_user(user, nav_to_user=False)
+            private = False
+            
+        # User is private
+        except PrivateAccountError:
+            private = True
 
-        self.nav_user(user)
+        if self.__check_existence(EC.presence_of_element_located((By.XPATH, Paths.REQUESTED_BTN))):
+            # Follow request already sent
+            pass
+        elif self.__check_existence(EC.presence_of_element_located((By.XPATH, Paths.MESSAGE_USER_BTN))):
+            # User already followed
+            pass
+        else:
+            follow_button = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.FOLLOW_BTN)), url=ClientUrls.NAV_USER.format(user))
+            follow_button.click()
 
-        follow_buttons = self.__find_buttons('Follow')
+        if discard_driver:
+            self.__discard_driver()
 
-        for btn in follow_buttons:
-            btn.click()
+        if private:
+            raise FollowRequestSentError(user)
 
     
     @insta_method
-    def unfollow_user(self, user:str):
+    def unfollow_user(self, user:str, discard_driver:bool=False):
         """
         Unfollows user(s)
 
         Args:
             user:str: Username of user to unfollow
         """
-
+        if not self.driver:
+            self.__init_driver(login=True)
         self.nav_user(user)
 
         unfollow_btns = self.__find_buttons('Following')
@@ -332,10 +366,12 @@ class InstaClient:
                 unfollow_confirmation.click()
         else:
             print('INSTACLIENT: No {} buttons were found.'.format('Following'))
+        if discard_driver:
+            self.__discard_driver()
     
 
     @insta_method
-    def get_user_images(self, user:str):
+    def get_user_images(self, user:str, discard_driver:bool=False):
         """
         Get all images from a users profile.
 
@@ -353,7 +389,7 @@ class InstaClient:
         finished = False
         while not finished:
 
-            finished = self._infinite_scroll() # scroll down
+            finished = self.__infinite_scroll() # scroll down
             
             elements = self.__find_element((EC.presence_of_element_located(By.CLASS_NAME, 'FFVAD')))
             img_srcs.extend([img.get_attribute('src') for img in elements]) # scrape srcs
@@ -363,7 +399,7 @@ class InstaClient:
     
 
     @insta_method
-    def like_latest_posts(self, user:str, n_posts:int, like:bool=True):
+    def like_latest_posts(self, user:str, n_posts:int, like:bool=True, discard_driver:bool=False):
         """
         Likes a number of a users latest posts, specified by n_posts.
 
@@ -396,7 +432,7 @@ class InstaClient:
 
 
     @insta_method
-    def send_dm(self, user:str, message:str, check_user=True):
+    def send_dm(self, user:str, message:str, discard_driver:bool=False):
         """
         Send an Instagram Direct Message to a user. if `check_user` is set to True, the `user` argument will be checked to validate whether it is a real instagram username.
 
@@ -405,13 +441,19 @@ class InstaClient:
             message (str): Message to send to the user via DMs
             check_user (bool, optional): If set to False, the `InstaClient` will assume that `user` is a valid instagram username. Defaults to True.
         """
+        if not self.driver:
+            self.__init_driver(login=True)
         # Navigate to User's dm page
-        self.nav_user_dm(user, check_user=check_user)
-        text_area = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.DM_TEXT_AREA)))
-        print(text_area)
-        text_area.send_keys(message)
-        send_btn = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.SEND_DM_BTN)))
-        send_btn.click()
+        try:
+            self.nav_user_dm(user)
+            text_area = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.DM_TEXT_AREA)))
+            text_area.send_keys(message)
+            send_btn = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.SEND_DM_BTN)))
+            send_btn.click()
+        except Exception as error: 
+            if discard_driver:
+                self.__discard_driver()
+            raise error
 
 
     #@insta_method
@@ -429,11 +471,23 @@ class InstaClient:
 
 
     @insta_method
-    def scrape_followers(self, user:str, check_user=True, callback=None):
+    def scrape_followers(self, user:str, check_user=True, discard_driver:bool=False):
+        """
+        scrape_followers Scrape an instagram user's followers and return them as a list of strings.
+
+        Args:
+            user (str): User to scrape
+            check_user (bool, optional): If set to True, checks if the `user` is a valid instagram username. Defaults to True.
+            discard_driver (bool, optional): If set to True, the `driver` will be discarded at the end of the method. Defaults to False.
+
+        Returns:
+            list: List of instagram usernames
+        """
         
+        if not self.driver:
+            self.__init_driver(login=True)
         # Nav to user page
         self.nav_user(user, check_user=check_user)
-        self.driver.save_screenshot('user.png') # TODO remove after debugging
         # Find Followers button/link
         followers_btn:WebElement = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.FOLLOWERS_BTN)), url=ClientUrls.NAV_USER.format(user))
         # Start scraping
@@ -441,7 +495,6 @@ class InstaClient:
         # Click followers btn
         followers_btn.click()
         time.sleep(2)
-        self.driver.save_screenshot('followers.png') # TODO remove after debugging
         # Load all followers
         followers = []
         main:WebElement = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.FOLLOWERS_LIST_MAIN)))
@@ -458,7 +511,11 @@ class InstaClient:
                 continue
             else:
                 break
-        followers_list:WebElement = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.FOLLOWERS_LIST)), wait_time=3)
+        try:
+            followers_list:WebElement = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.FOLLOWERS_LIST)), wait_time=3)
+        except NoSuchElementException:
+            return self.scrape_followers(user, check_user, discard_driver)
+
         divs = followers_list.find_elements_by_xpath(Paths.FOLLOWER_USER_DIV)
         for div in divs:
             try:
@@ -467,6 +524,8 @@ class InstaClient:
                     followers.append(username)
             except:
                 pass
+        if discard_driver:
+            self.__discard_driver()
         return followers
 
                 
@@ -488,7 +547,7 @@ class InstaClient:
     
 
     @insta_method
-    def search_tag(self, tag:str):
+    def search_tag(self, tag:str, discard_driver:bool=False):
         """
         Naviagtes to a search for posts with a specific tag on IG.
 
@@ -507,21 +566,24 @@ class InstaClient:
 
 
     @insta_method
-    def logout(self):
+    def logout(self, discard_driver:bool=False):
         """
         Check if the client is currently connected to Instagram and logs of the current InstaClient session.
 
         Returns:
-            [type]: True if the 
+            bool: True if the 
         """
         result = self.check_status()
         if result:
-            settings_btn = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.SETTINGS_BTN)), wait_time=4)
-            settings_btn.click()
-            logout_btn = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.LOG_OUT_BTN)), wait_time=4)
-            logout_btn.click()
-            confirm_btn = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.CONFIRM_LOGOUT_BTN)), wait_time=4)
-            confirm_btn.click()
+            if discard_driver:
+                self.__discard_driver()
+            else:
+                settings_btn = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.SETTINGS_BTN)), wait_time=4)
+                settings_btn.click()
+                logout_btn = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.LOG_OUT_BTN)), wait_time=4)
+                logout_btn.click()
+                confirm_btn = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.CONFIRM_LOGOUT_BTN)), wait_time=4)
+                confirm_btn.click()
             return True
         else:
             return True
@@ -562,33 +624,47 @@ class InstaClient:
             True if operation was successful
         """
         self.nav_user(user, check_user=check_user)
-        message_btn = self.__find_buttons('Message')
-        # Open User DM Page
-        message_btn.click()
-        return True
+        if self.__check_existence(EC.presence_of_element_located((By.XPATH, Paths.FOLLOW_BTN))):
+            self.follow_user(user, nav_to_user=False)
+        if self.__check_existence(EC.presence_of_element_located((By.XPATH, Paths.REQUESTED_BTN))):
+            raise FollowRequestSentError(user)
+        else:
+            message_btn = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.MESSAGE_USER_BTN)))
+            # Open User DM Page
+            message_btn.click()
+            return True
         
 
     @insta_method
-    def is_valid_user(self, user, nav_to_user=True):
+    def is_valid_user(self, user, nav_to_user=True, discard_driver:bool=False):
+        if not self.driver:
+            self.__init_driver(login=True)
+            nav_to_user = True
         if nav_to_user:
             self.driver.get(ClientUrls.NAV_USER.format(user))
         element = self.__check_existence(EC.presence_of_element_located((By.XPATH, Paths.PAGE_NOT_FOUND)), wait_time=8)
         if element:
             # User does not exist
             self.driver.get(ClientUrls.HOME_URL)
+            if discard_driver:
+                self.__discard_driver()
             raise InvalidUserError(username=user)
         else: 
             # Operation Successful
             paccount_alert = self.__check_existence(EC.presence_of_element_located((By.XPATH, Paths.PRIVATE_ACCOUNT_ALERT)), wait_time=3)
             if paccount_alert:
                 # navigate back to home page
+                if discard_driver:
+                    self.__discard_driver()
                 raise PrivateAccountError(user)
             else:
+                if discard_driver:
+                    self.__discard_driver()
                 return True
 
 
     # IG PRIVATE UTILITIES
-    def _infinite_scroll(self):
+    def __infinite_scroll(self):
         """
         Scrolls to the bottom of a users page to load all of their media
 
@@ -626,7 +702,7 @@ class InstaClient:
         return buttons
 
 
-    def __find_element(self, expectation, url:str=None, wait_time:int=15, attempt=0):
+    def __find_element(self, expectation, url:str=None, wait_time:int=8, attempt=0):
         """
         __find_element finds and returns the `WebElement`(s) that match the expectation's XPATH.
 
@@ -651,28 +727,19 @@ class InstaClient:
         except TimeoutException:
             # Element was not found in time
             if attempt < 1:
-                current_url = self.driver.current_url
                 if self.__check_existence(EC.presence_of_element_located((By.XPATH, Paths.COOKIES_LINK))):
                     accept_btn = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.ACCEPT_COOKIES)))
                     accept_btn.click()
                 if not self.check_status():
                     # Not Logged In!
                     raise NotLoggedInError()
-                if url is not None and url not in current_url:
-                    # Wrong page
-                    if attempt > 0:
-                        # Element not found
-                        self.driver.save_screenshot('element_not_found.png')
-                        raise NoSuchElementException()
-                    else:
+                else:
+                    if attempt < 1 and url is not None:
                         self.driver.get(url)
                         self.__find_element(self, expectation, url, wait_time, attempt+1)
-                else:
-                    # refresh page
-                    self.driver.navigate().refresh();
-                    self.__find_element(expectation, url, wait_time, attempt+1)
+                    else:
+                        raise NoSuchElementException()
             else:
-                self.driver.save_screenshot('element_not_found.png')
                 raise NoSuchElementException()
 
 
@@ -689,4 +756,48 @@ class InstaClient:
             return True
         except:
             return False
+
+
+    def __discard_driver(self):
+        if self.driver:
+            self.driver.quit()
+            self.logged_in = False
+            self.driver = None
+
+
+    def __init_driver(self, login=False):
+        try:
+            if self.driver_type == self.CHROMEDRIVER:
+                if self.host_type == self.WEB_SERVER:
+                    # Running on web server
+                    chrome_options = webdriver.ChromeOptions()
+                    chrome_options.binary_location = os.environ.get("GOOGLE_CHROME_BIN")
+                    chrome_options.add_argument('--user-agent=Mozilla/5.0 (iPhone; CPU iPhone OS 10_3 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) CriOS/56.0.2924.75 Mobile/14E5239e Safari/602.1')
+                    chrome_options.add_argument("--window-size=343,915")
+                    chrome_options.add_argument("--headless")
+                    chrome_options.add_argument("--disable-dev-shm-usage")
+                    chrome_options.add_argument("--no-sandbox")
+                    self.driver = webdriver.Chrome(executable_path=os.environ.get("CHROMEDRIVER_PATH"), chrome_options=chrome_options)
+                elif self.host_type == self.LOCAHOST:
+                    # Running locally
+                    chrome_options = webdriver.ChromeOptions()
+                    chrome_options.add_argument('--user-agent=Mozilla/5.0 (iPhone; CPU iPhone OS 10_3 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) CriOS/56.0.2924.75 Mobile/14E5239e Safari/602.1')
+                    chrome_options.add_argument("--window-size=343,915")
+                    #chrome_options.add_argument("--headless")
+                    chrome_options.add_argument("--disable-dev-shm-usage")
+                    chrome_options.add_argument("--no-sandbox")
+                    self.driver = webdriver.Chrome(executable_path='instaclient/drivers/chromedriver.exe', chrome_options=chrome_options)
+                else:
+                    raise InvaildHostError(self.host_type)
+            else:
+                raise InvaildDriverError(self.driver_type)
+        except Exception as error:
+            raise error
+
+        if login:
+            try:
+                self.login(self.username, self.password)
+            except:
+                raise InstaClientError(message='Tried logging in when initiating driver, but username and password are not defined.')
+
         
\ No newline at end of file
